import logging

# json keys in orthanc response
TYPE =                          'Type'
ID =                            'ID'
PARENT  =                       'Parent'

UNKNOWN =                       '!UNKNOWN!'
LOG_LEVEL =                     logging.ERROR
PRIVATE =                       'private'
ASCENDING =                     'ascending'
DESCENDING =                    'descendig'

# when requesting more than this number of items a progressbar will be shown
SHOW_PROGRESS_BAR_COUNT =       1000
